/* Copyright (C) 2000 KOM/Darmstadt University of Technology
 *
 * You are allowed to use all other parts of the code under the following terms:
 *
 * For non-commercial use, code may be used in unmodified form provided
 * that this copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranty of
 * fitness for a particular purpose.
 *
 * The code may be subjected to the GNU General Public License, Version 2,
 * and re-distributed under the terms of this license.
 *
 * Commercial use other than under the terms of the GNU General Public
 * License is allowed only after express negotiation of conditions
 * with the authors.
 */

#include "Proxy.h"
#include "Session.h"
#include "Log.h"
#include "Conf.h"
#include "SignalHandler.h"
#include "TimeValue.h"

#include <signal.h>     /* signals */
#include <errno.h>      /* errno */
#include <unistd.h>	/* pipe, close, read */
#include <string.h>	/* bzero */

#define max(i,j)        (((i)>(j)) ? (i) : (j))

Proxy::Proxy():maxFD(0),size(0),sizeFD(0){
	TimeValue::getCurrentSystemTime(currTime);
	FD_ZERO(&setFD);
	FD_ZERO(&workSetFD);
        for(int i=0;i<MAXSESSIONS;++i){
                sessionList[i]=NULL;
        }
	pipe(transferPipe);
	addFD(transferPipe[0]);
	INF((__FILE__,__FUNCTION__,"proxy started %s\n",getDescript().c_str()))
}

Proxy::~Proxy() {
	this->stop();
	INF((__FILE__,__FUNCTION__,"waiting for proxy thread to terminate %s\n",getDescript().c_str()))
	join();
	delSession();
	close(transferPipe[0]);
	close(transferPipe[1]);
	INF((__FILE__,__FUNCTION__,"proxy terminated %s\n",getDescript().c_str()))
}

int Proxy::getTransferFD(){
	return transferPipe[1];
}

String Proxy::getDescript(){
        /* [thread: 1025 (sessions:20|FDs:0)] */
        String d;

        d = d + "[thread: " 	+ myID();
        d = d + " (sessions:"  	+ size; 
        d = d + "|FDs:"     	+ sizeFD;
        d = d + ")]";

        return (d);
}

int Proxy::addSession(Session* s){
        for(int i=0;i<MAXSESSIONS;++i){
                if (sessionList[i]==NULL){
                        s->setProxy(this);
			s->setSessionID(i);
                        sessionList[i]=s;
			LOG((__FILE__,__FUNCTION__,4,"add session %s\n",sessionList[i]->getDescript().c_str()))
                        ++size;
                        return(0);
                }
        }
        WRN((__FILE__,__FUNCTION__,"proxy overloaded, could not handle new session %s\n",getDescript().c_str()))
        return(-1);
}

int Proxy::delSession(int id){
	if ((id<0)||(id>MAXSESSIONS)) FAT((__FILE__,__FUNCTION__,"session id out of scope: %d %s\n",id,getDescript().c_str()))
        if (sessionList[id]!=NULL){
		LOG((__FILE__,__FUNCTION__,4,"del session %s\n",sessionList[id]->getDescript().c_str()))
		delete (sessionList[id]);
		sessionList[id]=NULL;
		--size;
		return(0);
	}
        FAT((__FILE__,__FUNCTION__,"could not find session in list %s\n",getDescript().c_str()))
}

int Proxy::delSession(Session* s){
	return delSession(s->getSessionID());
}

int Proxy::delSession(){
        for(int i=0;i<MAXSESSIONS;++i){
                if (sessionList[i]!=NULL){
			LOG((__FILE__,__FUNCTION__,4,"del session %s\n",sessionList[i]->getDescript().c_str()))
                        delete sessionList[i];
                        sessionList[i]=NULL;
                }
        }
        size=0;
        return(0);
}

int Proxy::getSize(){
	return(size);
}

bool Proxy::hasSpaceLeft(){
	if (size<(MAXSESSIONS-1)) return (true);
	return(false);
}

int Proxy::proxyFinish(){

	TimeValue::getCurrentSystemTime(currTime);

        for(int i=0;i<MAXSESSIONS;++i){
                if (sessionList[i]!=NULL){
                        if ((sessionList[i]->getSessionEnd())==true) {
                                delSession(sessionList[i]);
                        } 
			else { 
				if ((sessionList[i]->getTimeoutTime())!=TimeValue(0,0)){
                        		if (currTime > ((sessionList[i]->getStartTime())+(sessionList[i]->getTimeoutTime()))) {
        					INF((__FILE__,__FUNCTION__,"session timed out %s\n",getDescript().c_str()))
                                		delSession(sessionList[i]);
                        		} 
				}
			}
                }
        }
        return(0);
}

int Proxy::proxySelect(){
	TimeValue t=TimeValue(1,0);
        int ret=0;

        /* copy the stored fdset */
        memcpy(&workSetFD,&setFD,sizeof(workSetFD));

        ret=select(maxFD,&workSetFD, (fd_set *)NULL, (fd_set *)NULL,(struct timeval*) &t);

        if ((ret<0) && (errno==EINTR)) return(-3);
        else if (ret==0) return(-4);
        else if (ret<0)  return(-1);
        return(ret);
}

int Proxy::proxyInitSpecial(){
        return(0);
}

int Proxy::proxyInitSession(){
        return(0);
}

int Proxy::proxyInit(){
        int ret;

        if ((ret=(proxyInitSpecial()))<0) return ret;
        if ((ret=(proxyInitSession()))<0) return ret;
        return(0);
}


int Proxy::proxyRelaySpecialFD(){
	if(FD_ISSET(transferPipe[0],&workSetFD)){
                Session* s;
                if (read(transferPipe[0],&s,sizeof(s))!=sizeof(s)){
                        ERR((__FILE__,__FUNCTION__,"reading session pointer %s\n",getDescript().c_str()))
                        return(-1);
                }
                if (addSession(s)<0){
                        ERR((__FILE__,__FUNCTION__,"queuing session in proxy %s\n",getDescript().c_str()))
                        delete(s);
                        return(0);
                }
                return(0);
        }
	return(1);
}

int Proxy::proxyRelaySessionFD(){
        int ret;

        for(int i=0;i<MAXSESSIONS;++i){
                if (sessionList[i]!=NULL) sessionList[i]->sessionRead(&workSetFD);
        }
        return(0);
}


int Proxy::proxyRelay(){
        int ret;

        if ((ret=(proxyRelaySpecialFD()))<=0) return ret;
        if ((ret=(proxyRelaySessionFD()))<=0) return ret;
        return(0);
}

int Proxy::log(int loglevel){
        LOG((__FILE__,__FUNCTION__,loglevel,"proxy                 thread# %d sessions# %d/%d FD# %d\n",myID(),getSize(),MAXSESSIONS,sizeFD))
        return(0);
}

void Proxy::main() {
	int ret;

	LOG((__FILE__,__FUNCTION__,10,"starting Proxy loop %s\n",getDescript().c_str()))
        SignalHandler::unblock(SIGUSR1);
        SignalHandler::unblock(SIGUSR2);
	/* init all sessions */
        if ((ret=proxyInit())<0){
                ERR((__FILE__,__FUNCTION__,"proxyInit %s\n",getDescript().c_str()))
                end=true;
        }
	while ( !end ) {
		/* clean up all finished sessions */
                if ((ret=proxyFinish())<0){
                        ERR((__FILE__,__FUNCTION__,"proxyFinish %s\n",getDescript().c_str()))
                        end=true;
                }
		
                if ((ret=proxySelect())<0){
                        if (ret==-1) {
                                ERR((__FILE__,__FUNCTION__,"proxySelect %s\n",getDescript().c_str()))
                        	this->stop();
                        }
                        else if (ret==-2) {
                                LOG((__FILE__,__FUNCTION__,10,"no fds in select %s\n",getDescript().c_str()))
                        }
                        else if (ret==-3) {
                                LOG((__FILE__,__FUNCTION__,10,"select interrupted %s\n",getDescript().c_str()))
                        }
                        else if (ret==-4) {
                                LOG((__FILE__,__FUNCTION__,40,"select timeout %s\n",getDescript().c_str()))
                        }
                }
                /* if session data is available process session data */
                else {
                        if ((ret=proxyRelay())<0){
                                ERR((__FILE__,__FUNCTION__,"proxyRelay %s\n",getDescript().c_str()))
                        	this->stop();
                        }
                }
	}
	LOG((__FILE__,__FUNCTION__,10,"leaving Proxy loop %s\n",getDescript().c_str()))
}

fd_set* Proxy::getSetFD(){
        return (&setFD);
}

int Proxy::getMaxFD(){
        return (maxFD);
}

void Proxy::addFD(int fd){
        FD_SET(fd,&setFD);
        maxFD = max((maxFD-1),fd);
        ++maxFD;
	++sizeFD;
}

void Proxy::delFD(int fd){
        FD_CLR(fd,&setFD);
	--sizeFD;
}

